from openai import OpenAI

from .card import Card
from .deck import Deck
from .config import Config

class TarotReading:
    def __init__(self, tarot_cards: list[Card], question_asked: str, response: str):
        self.__tarot_cards = tarot_cards
        self.__question_asked = question_asked
        self.__response = response

    @property
    def tarot_cards(self) -> list[Card]:
        return self.__tarot_cards

    @property
    def question_asked(self) -> str:
        return self.__question_asked

    @property
    def response(self) -> str:
        return self.__response

class TarotReader:
    __init_question: str = 'You are a intelligent assistant who specializes in tarot card readings.'

    def __init__(self, open_ai_api_key: str, model_name: str):
        self.config = Config()
        self.__messages: list[dict[str, str]] = list()
        self.__model_name: str = model_name
        self.__open_ai_api_key: str = open_ai_api_key

        self.__client = OpenAI(api_key=self.__open_ai_api_key)
        self.__write_using_stream = self.config.chatgpt_write_using_stream

    def set_ai_model(self, model_name: str) -> None:
        self.__model_name = model_name

    def get_reading(self, question: str, additional_context: list[str] = None, num_cards: int = 3) -> TarotReading:
        deck = Deck()
        cards_drawn: list[Card] = list()

        for i in range(num_cards):
            cards_drawn.append(deck.draw_card())

        self.__load_client()

        # Initial question
        self.__append_user_msg(question)

        # Context
        if additional_context:
            self.__append_user_msg('The next messages provide additional context to enhance your reading so that it is '
                                   'tailored to this context.')
            for context in additional_context:
                self.__append_user_msg(context)

        card_string = ', '.join(
            f'{card} in {"reversed" if card.is_reversed else "upright"} position' for card in cards_drawn
        )

        self.__append_user_msg(f'The following cards are drawn: {card_string}')

        chat_completion = self.__client.chat.completions.create(
            model=self.__model_name, messages=self.__messages, stream=self.__write_using_stream
        )

        assistant_reply = chat_completion.choices[0].message.content

        self.__append_user_msg(assistant_reply)

        return TarotReading(
            tarot_cards=cards_drawn,
            question_asked=question,
            response=assistant_reply
        )

    def __load_client(self):
        self.__start_new_thread()

    def __append_user_msg(self, message: str) -> None:
        self.__messages.append({'role': 'user',
                                'content': message})

    def __append_assistant_msg(self, message: str) -> None:
        self.__messages.append({'role': 'assistant',
                                'content': message})

    def __append_system_msg(self, message: str) -> None:
        self.__messages.append({'role': 'system',
                                'content': message})

    def __start_new_thread(self):
        self.__clear_messages()
        self.__append_system_msg(self.__init_question)

    def __clear_messages(self) -> None:
        self.__messages.clear()



import os
import configparser

class Config:
    def __init__(self):
        self.__config_file: str = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'config.ini')
        self.__config = configparser.ConfigParser()
        self.__config.read(self.__config_file)

    @property
    def reverse_probability(self) -> float:
        return self.__config.getfloat('Tarot', 'reverse_probability')

    @reverse_probability.setter
    def reverse_probability(self, value: float):
        if value < 0 or value > 1:
            raise ValueError("Reverse probability must be between 0 and 1")
        self.__write_to_config('Tarot', 'reverse_probability', value)

    def __write_to_config(self, category, variable, value):
        valstr = str(value)
        self.__config.set(category, variable, valstr)
        with open(self.__config_file, 'w') as configfile:
            self.__config.write(configfile)

    @property
    def chatgpt_write_using_stream(self):
        return self.__config.getboolean('ChatGPT', 'write_using_stream')

    @chatgpt_write_using_stream.setter
    def chatgpt_write_using_stream(self, value: bool):
        self.__config.set('ChatGPT', 'write_using_stream', str(value))


class NoCardsRemainingError(Exception):
    def __init__(self):
        self.message = 'Attempted to draw a card from an empty deck.'
        super().__init__(self.message)

class UnexpectedDirectionError(Exception):
    def __init__(self, direction: str):
        self.message = (f'The inputted direction ("{direction}") is not a recognized value. '
                        f'Please input either "upright" or "reversed".')
        super().__init__(self.message)


import random
from dataclasses import dataclass
from typing import Self

from .config import Config
from .exceptions import UnexpectedDirectionError


@dataclass
class CardRaw:
    suit: str
    name: str
    upright_meanings_1: list[str]
    upright_meanings_2: list[str]
    upright_love_meaning: list[str]
    upright_career_meaning: list[str]
    upright_finances_meaning: list[str]
    reverse_meanings_1: list[str]
    reverse_meanings_2: list[str]
    reverse_love_meaning: list[str]
    reverse_career_meaning: list[str]
    reverse_finances_meaning: list[str]


class Card:
    def __init__(self, raw_card: CardRaw, forced_direction: str = 'none'):
        self.config = Config()
        self.raw_card = raw_card
        self.__forcedDirection = forced_direction

        self.is_reversed: bool = self.__set_reversed_status()

    @property
    def name(self) -> str: return self.raw_card.name

    @property
    def suit(self) -> str: return self.raw_card.suit

    @property
    def meanings_summary1(self) -> list[str]:
        if self.is_reversed:
            return self.raw_card.reverse_meanings_1
        return self.raw_card.upright_meanings_1

    @property
    def meanings_summary2(self) -> list[str]:
        if self.is_reversed:
            return self.raw_card.reverse_meanings_2
        return self.raw_card.upright_meanings_2

    @property
    def meanings_love(self) -> list[str]:
        if self.is_reversed:
            return self.raw_card.reverse_love_meaning
        return self.raw_card.upright_love_meaning

    @property
    def meanings_career(self) -> list[str]:
        if self.is_reversed:
            return self.raw_card.reverse_career_meaning
        return self.raw_card.upright_career_meaning

    @property
    def meanings_finances(self) -> list[str]:
        if self.is_reversed:
            return self.raw_card.reverse_finances_meaning
        return self.raw_card.upright_finances_meaning

    def __set_reversed_status(self) -> bool:
        if self.__forcedDirection == 'none':
            if random.random() < self.config.reverse_probability:
                return True
            return False
        else:
            if self.__forcedDirection.lower() == 'upright':
                return False
            elif self.__forcedDirection.lower() == 'reversed':
                return True
            else:
                raise UnexpectedDirectionError(self.__forcedDirection)

    def get_card_in_other_direction(self) -> Self:
        direction = 'upright' if self.is_reversed else 'reversed'
        return Card(raw_card=self.raw_card, forced_direction=direction)

    def __str__(self):
        return self.name


import json
import os
import random
from argparse import ArgumentError

from .config import Config
from .card import Card, CardRaw
from .exceptions import NoCardsRemainingError


class Deck:
    __card_directory = os.path.join(os.path.dirname(os.path.realpath(__file__)), 'cards')

    def __init__(self):
        self.config = Config()
        self.__cards: list[CardRaw] = list()
        self.reset()

    @property
    def count(self) -> int:
        return len(self.__cards)

    def draw_card(self, force_upright: bool = False, force_reverse: bool = False) -> Card:
        if self.count == 0:
            raise NoCardsRemainingError()

        if force_upright and force_reverse:
            raise ArgumentError(message='A card cannot be forced to be drawn in both upright and reverse positions.')

        raw_card = self.__cards.pop(random.randrange(self.count))

        if force_upright:
            return Card(raw_card, 'upright')

        if force_reverse:
            return Card(raw_card, 'reversed')

        return Card(raw_card)

    def reset(self) -> None:
        self.__cards = list()
        self.__load_deck()

    def __load_deck(self) -> None:
        for file in os.listdir(self.__card_directory):
            if file.endswith('.json'):
                with open(os.path.join(self.__card_directory, file), 'r') as f:
                    self.__cards.append(CardRaw(**json.load(f)))
